from setuptools import setup, find_packages

version = '0.1.0a2'

setup(
    name='mentpy',
    version=version,
    author='Luis Mantilla',
    author_email='luismantilla99@outlook.com',
    description='A Python library for simulating MBQC circuits',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/BestQuark/mentpy',
    packages=find_packages(),
)
